__all__ = ["coreset", "sparse_gmm", "imputation_util", "nonparanormal", "json_util"]
from coreset import get_coreset
from imputation_util import predict_missing_values, sample_missing_values
from nonparanormal import NPNTransformer
from sparse_gmm import SparseGMM
from json_util import GamelanPyEncoder, gamelan_json_obj_hook



#!/usr/bin/python
__author__ = 'jiohyoo'
import sys
import gamelanpy
import getopt
import numpy as np
import json

def getopts(argv):
    try:
        opts, args = getopt.getopt(argv, "h", [
            "help", "n-samples="])
    except getopt.GetoptError:
        usage()
        exit(2)

    method = 'predict'
    n_samples = None
    for opt, arg in opts:
        print opt, arg
        if opt in ['-h', '--help']:
            usage()
            exit()
        elif opt == '--n-samples':
            method = 'sample'
            n_samples = int(arg)

    if method == 'sample' and n_samples <= 0:
        print 'Specified number of samples: %d must be positive integer' % n_samples
        exit(2)

    if method not in ['predict', 'sample']:
        print 'Specified method: %s is invalid' % method
        exit(2)

    return method, n_samples


def main(model, data_path, output_path, method, n_samples):

    data = np.genfromtxt(data_path, delimiter=',');
    if method == 'predict':
        output = gamelanpy.predict_missing_values(model, data)
    elif method == 'sample':
        output = gamelanpy.sample_missing_values(model, data, n_samples=n_samples)

    np.savetxt(output_path, output, delimiter=',')


def usage():
    print 'print usage'

if __name__ == '__main__':
    model = json.load(open(sys.argv[1]), object_hook=gamelanpy.gamelan_json_obj_hook)
    data_path = sys.argv[2]
    output_path = sys.argv[3]
    method, n_samples = getopts(sys.argv[4:])
    main(model, data_path, output_path, method, n_samples)






import numpy as np
import scipy
import scipy.linalg
import scipy.cluster
from sklearn.mixture import GMM
from coreset import get_coreset
import nonparanormal

from sklearn.covariance import graph_lasso
from weighted_kmeans import WeightedKMeans


def _learn_sparse_gaussian(data, l1_penalty_range, l1_search_depth, l1_search_repeat, atol):
    '''
    Learn single sparse Gaussian Graphical Model using graphical lasso
    L1-parameter search is done by heuristic method using column shuffling

    :param data: array of shape(n_samples, n_dimensions)
    :param l1_penalty_range: min and max values for l1_penalty search
    :param l1_search_depth: number of depth in binary search for l1_penalty
    :param l1_search_repeat: number of trials for l1_penalty search
    :param atol: absolute tolerance for recovering zero or non-zero entries in precision matrix
    :return: covarinace matrix, precision matrix, final l1_penalty
    '''
    n_frames, n_vars = data.shape

    # copy data for column shuffling
    data_s = data.copy()
    l1_penalty_sum = 0.
    for i in range(l1_search_repeat):
        l1_min, l1_max = l1_penalty_range
        l1_mid = (l1_min + l1_max) / 2
        for j in range(l1_search_depth):
            # shuffle columns
            for p in range(n_vars):
                data_s[:, p] = data_s[np.random.permutation(n_frames), p]
            cov, pre = graph_lasso(np.cov(data_s.T), l1_mid)

            # conservative binary search on l1_param
            if ((np.isclose(pre, 0, atol=atol) == False) == np.eye(n_vars)).all():
                l1_max = (l1_max + l1_mid) / 2
            else:
                l1_min = (l1_min + l1_mid) / 2
            l1_mid = (l1_min + l1_max) / 2
        l1_penalty_sum += l1_mid

    l1_penalty = l1_penalty_sum / l1_search_repeat

    # final graphical lasso on the original data
    cov, pre = graph_lasso(np.cov(data.T), l1_penalty)

    pre[np.abs(pre) < atol] = 0.
    cov = np.linalg.inv(pre)
    return cov, pre, l1_penalty


class SparseGMM(GMM):
    def __init__(self, n_components=1, nonparanormal=False, random_state=None, thresh=None, min_covar=0.001,
                 n_iter=100, n_init = 1, params='wmc', init_params='wmc', tol=1e-4):
        GMM.__init__(self, n_components=n_components, covariance_type='full', random_state=random_state,
                     thresh=thresh, min_covar=min_covar, n_iter=n_iter, n_init=n_init, params=params,
                     init_params=init_params)
        self.nonparanormal = nonparanormal
        self.tol=tol
        self.l1_penalty_ = None

    def bic(self, X):
        """Bayesian information criterion for the current model fit
        and the proposed data

        Parameters
        ----------
        X : array of shape(n_samples, n_dimensions)

        Returns
        -------
        bic: float (the greater the better)
        """
        n_vars = self.means_.shape[1]
        score = self.score(X).sum()
        df = 0
        for k in range(self.n_components):
            df += (np.abs(scipy.triu(self.precs_[k])) > self.tol).sum()
            df += n_vars

        return score - df * np.log(X.shape[0]) / 2

    def aic(self, X):
        """Akaike information criterion for the current model fit
        and the proposed data

        Parameters
        ----------
        X : array of shape(n_samples, n_dimensions)

        Returns
        -------
        aic: float (the greater the better)
        """
        n_vars = self.means_.shape[1]
        score = self.score(X).sum()
        df = 0
        for k in range(self.n_components):
            df += (np.abs(scipy.triu(self.precs_[k])) > self.tol).sum()
            df += n_vars

        return score - df

    def fit(self, X, subsample_method, subsample_size,
            l1_penalty_range=[0.0001, 10.0], l1_search_depth=20, l1_search_repeat=10, npn_sample_ratio=None):
        n_frames, n_vars = X.shape

        samples = None
        sample_idx = None

        if subsample_method == 'coreset':
            coreset, weights = get_coreset(X, self.n_components, subsample_size)
            kmeans = WeightedKMeans(n_clusters=self.n_components)
            kmeans.fit(coreset, weights=weights)
            X_idx = kmeans.predict(X)

            samples = X
            sample_idx = X_idx
        elif subsample_method == 'coreset2':
            coreset, weights = get_coreset(X, self.n_components, subsample_size)
            kmeans = WeightedKMeans(n_clusters=self.n_components)
            kmeans.fit(coreset, weights=weights)
            coreset_idx = kmeans.predict(coreset)

            samples = coreset
            sample_idx = coreset_idx

        elif subsample_method == 'uniform':
            subsamples = X[np.random.permutation(n_frames)[:subsample_size], :]
            kmeans = WeightedKMeans(n_clusters=self.n_components)
            kmeans.fit(subsamples)

            samples = X
            sample_idx = kmeans.predict(X)
        elif subsample_method == 'None':
            kmeans = WeightedKMeans(n_clusters=self.n_components)
            kmeans.fit(X)
            samples = X
            sample_idx = kmeans.predict(X)
        else:
            raise ValueError('Method %s is not supported' % subsample_method)

        self.covars_ = np.empty((self.n_components, n_vars, n_vars), dtype=float)
        self.precs_ = np.empty((self.n_components, n_vars, n_vars), dtype=float)
        self.means_ = np.empty((self.n_components, n_vars), dtype=float)
        self.weights_ = np.empty(self.n_components, dtype=float)
        self.l1_penalty_ = np.empty(self.n_components, dtype=float)
        if self.nonparanormal:
            self.npns_ = [None] * self.n_components

        for k in range(self.n_components):
            data_k = samples[sample_idx == k, :]

            if self.nonparanormal:
                self.npns_[k] = nonparanormal.NPNTransformer()
                self.npns_[k].set_data(data_k, ratio=npn_sample_ratio)
                data_k = self.npns_[k].transform(data_k)

            self.means_[k] = data_k.mean(axis=0)

            cov, pre, l1_penalty = \
                _learn_sparse_gaussian(X, l1_penalty_range, l1_search_depth, l1_search_repeat, self.tol)
            self.covars_[k] = cov
            self.precs_[k] = pre
            self.l1_penalty_[k] = l1_penalty
            self.weights_[k] = data_k.shape[0]

        self.weights_ /= self.weights_.sum()










# Original Author: Gael Varoquaux <gael.varoquaux@normalesup.org>
#          Thomas Rueckstiess <ruecksti@in.tum.de>
#          James Bergstra <james.bergstra@umontreal.ca>
#          Jan Schlueter <scikit-learn@jan-schlueter.de>
#          Nelle Varoquaux
#          Peter Prettenhofer <peter.prettenhofer@gmail.com>
#          Olivier Grisel <olivier.grisel@ensta.org>
#          Mathieu Blondel <mathieu@mblondel.org>
#          Robert Layton <robertlayton@gmail.com>
#
# Edited to cover weighted K-means algorithm by Ji Oh Yoo <jioh.yoo@gmail.com>
# License: BSD 3 clause

import numpy as np
from numpy import random
from sklearn.metrics.pairwise import euclidean_distances

import _weighted_kmeans


def weighted_kmeans(X, n_clusters, weights=None, n_init=10, max_iter=300, tol=1e-4):
    """Weighted K-means clustering algorithm.
    Parameters
    ----------
    X : array-like or sparse matrix, shape (n_samples, n_features)
        The observations to cluster.

    n_clusters : int
        The number of clusters to form as well as the number of
        centroids to generate.

    weights : array-like, shape(n_samples,), optional, default=None
        Weights for the given data

    max_iter : int, optional, default 300
        Maximum number of iterations of the k-means algorithm to run.

    n_init : int, optional, default: 10
        Number of time the k-means algorithm will be run with different
        centroid seeds. The final results will be the best output of
        n_init consecutive runs in terms of inertia.

    tol : float, optional
        The relative increment in the results before declaring convergence.


    Returns
    -------
    centroid : float ndarray with shape (k, n_features)
        Centroids found at the last iteration of k-means.

    label : integer ndarray with shape (n_samples,)
        label[i] is the code or index of the centroid the
        i'th observation is closest to.

    inertia : float
        The final value of the inertia criterion (sum of squared distances to
        the closest centroid for all observations in the training set).
    """
    if len(X.shape) == 1:
        X = X[:, np.newaxis]

    if n_init <= 0:
        raise ValueError("Number of iteration n_init=%d must be bigger than zero." % n_init)
    if n_clusters < 0:
        raise ValueError("Number of clusters n_clusters=%d must be at least 1." % n_clusters)
    if n_clusters > X.shape[0]:
        raise ValueError("Number of clusters n_clusters=%d is larger than number of samples %d"
                         % (n_clusters, X.shape[0]))

    X = np.array(X, dtype=float)
    tol = _tolerance(X, tol)

    # subtract mean of X for more accurate distance computations
    X_mean = X.mean(axis=0)
    X -= X_mean

    # precompute squared norms of data points
    x_squared_norms = np.einsum('ij,ij->i', X, X)

    best_labels, best_inertia, best_centers = None, None, None
    # For a single thread, less memory is needed if we just store one set
    # of the best results (as opposed to one set per run per thread).
    for it in range(n_init):
        # run a k-means once
        labels, inertia, centers, n_iter_ = _weighted_kmeans_single(
            X, n_clusters, weights=weights, max_iter=max_iter, tol=tol,
            x_squared_norms=x_squared_norms)
        # determine if these results are the best so far
        if best_inertia is None or inertia < best_inertia:
            best_labels = labels.copy()
            best_centers = centers.copy()
            best_inertia = inertia

    # add mean of X to original distribution
    best_centers += X_mean
    return best_centers, best_labels, best_inertia


def _weighted_kmeans_single(X, n_clusters, x_squared_norms, weights=None, max_iter=300, tol=1e-4):
    """A single run of k-means, assumes preparation completed prior.

    Parameters
    ----------
    X: array-like of floats, shape (n_samples, n_features)
        The observations to cluster.

    n_clusters: int
        The number of clusters to form as well as the number of
        centroids to generate.

    weights : array-like, shape(n_samples,), optional, default=None
        Weights for the given data

    max_iter: int, optional, default 300
        Maximum number of iterations of the k-means algorithm to run.

    tol: float, optional
        The relative increment in the results before declaring convergence.

    x_squared_norms: array
        Precomputed x_squared_norms.

    Returns
    -------
    centroid: float ndarray with shape (k, n_features)
        Centroids found at the last iteration of k-means.

    label: integer ndarray with shape (n_samples,)
        label[i] is the code or index of the centroid the
        i'th observation is closest to.

    inertia: float
        The final value of the inertia criterion (sum of squared distances to
        the closest centroid for all observations in the training set).

    n_iter : int
        Number of iterations run.
    """
    best_labels, best_inertia, best_centers = None, None, None
    # init
    centers = _k_init(X, n_clusters, x_squared_norms=x_squared_norms)

    # Allocate memory to store the distances for each sample to its
    # closer center for reallocation in case of ties
    distances = np.zeros(shape=(X.shape[0],), dtype=np.float64)

    # iterations
    for i in range(max_iter):
        centers_old = centers.copy()
        # labels assignment is also called the E-step of EM
        labels, inertia = \
            _labels_inertia(X, x_squared_norms, centers, weights=weights,
                            distances=distances)

        # computation of the means is also called the M-step of EM
        centers = _weighted_kmeans._centers_dense(X, labels, n_clusters, distances, weights)

        if best_inertia is None or inertia < best_inertia:
            best_labels = labels.copy()
            best_centers = centers.copy()
            best_inertia = inertia

        # break if diff is less than tol at this iteration
        diff = centers_old - centers
        if (diff * diff).sum() <= tol:
            break
    return best_labels, best_inertia, best_centers, i + 1


def _labels_inertia_precompute_dense(X, x_squared_norms, centers, distances, weights=None):
    """Compute labels and inertia using a full distance matrix.

    This will overwrite the 'distances' array in-place.

    Parameters
    ----------
    X : numpy array, shape (n_sample, n_features)
        Input data.

    x_squared_norms : numpy array, shape (n_samples,)
        Precomputed squared norms of X.

    centers : numpy array, shape (n_clusters, n_features)
        Cluster centers which data is assigned to.

    distances : numpy array, shape (n_samples,)
        Pre-allocated array in which distances are stored.

    weights : (optional) numpy array, shape (n_samples,)
        Weights for each point in X. Default is None

    Returns
    -------
    labels : numpy array, dtype=np.int, shape (n_samples,)
        Indices of clusters that samples are assigned to.

    inertia : float
        Sum of distances of samples to their closest cluster center.

    """
    n_samples = X.shape[0]
    k = centers.shape[0]
    all_distances = euclidean_distances(centers, X, x_squared_norms,
                                        squared=True)
    labels = np.empty(n_samples, dtype=np.int32)
    labels.fill(-1)
    mindist = np.empty(n_samples)
    mindist.fill(np.infty)
    for center_id in range(k):
        dist = all_distances[center_id]
        labels[dist < mindist] = center_id
        mindist = np.minimum(dist, mindist)
    if n_samples == distances.shape[0]:
        # distances will be changed in-place
        distances[:] = mindist
    if weights is None:
        inertia = mindist.sum()
    else:
        inertia = (mindist * weights).sum()
    return labels, inertia


def _labels_inertia(X, x_squared_norms, centers, weights=None, distances=None):
    """E step of the K-means EM algorithm.

    Compute the labels and the inertia of the given samples and centers.
    This will compute the distances in-place.

    Parameters
    ----------
    X: float64 array-like or CSR sparse matrix, shape (n_samples, n_features)
        The input samples to assign to the labels.

    x_squared_norms: array, shape (n_samples,)
        Precomputed squared euclidean norm of each data point, to speed up
        computations.

    centers: float64 array, shape (k, n_features)
        The cluster centers.

    weights : array-like, shape(n_samples,), optional, default=None
        Weights for the given data

    distances: float64 array, shape (n_samples,)
        Pre-allocated array to be filled in with each sample's distance
        to the closest center.

    Returns
    -------
    labels: int array of shape(n)
        The resulting assignment

    inertia : float
        Sum of distances of samples to their closest cluster center.
    """
    n_samples = X.shape[0]
    # set the default value of centers to -1 to be able to detect any anomaly
    # easily
    labels = -np.ones(n_samples, np.int32)
    if distances is None:
        distances = np.zeros(shape=(0,), dtype=np.float64)
    # distances will be changed in-place
    return _labels_inertia_precompute_dense(X, x_squared_norms, centers, distances, weights=weights)


def _k_init(X, n_clusters, x_squared_norms, n_local_trials=None):
    """Init n_clusters seeds according to k-means++

    Parameters
    -----------
    X: array or sparse matrix, shape (n_samples, n_features)
        The data to pick seeds for. To avoid memory copy, the input data
        should be double precision (dtype=np.float64).

    n_clusters: integer
        The number of seeds to choose

    x_squared_norms: array, shape (n_samples,)
        Squared Euclidean norm of each data point.

    random_state: numpy.RandomState
        The generator used to initialize the centers.

    n_local_trials: integer, optional
        The number of seeding trials for each center (except the first),
        of which the one reducing inertia the most is greedily chosen.
        Set to None to make the number of trials depend logarithmically
        on the number of seeds (2+log(k)); this is the default.

    Notes
    -----
    Selects initial cluster centers for k-mean clustering in a smart way
    to speed up convergence. see: Arthur, D. and Vassilvitskii, S.
    "k-means++: the advantages of careful seeding". ACM-SIAM symposium
    on Discrete algorithms. 2007

    Version ported from http://www.stanford.edu/~darthur/kMeansppTest.zip,
    which is the implementation used in the aforementioned paper.
    """
    n_samples, n_features = X.shape

    centers = np.empty((n_clusters, n_features))

    assert x_squared_norms is not None, 'x_squared_norms None in _k_init'

    # Set the number of local seeding trials if none is given
    if n_local_trials is None:
        # This is what Arthur/Vassilvitskii tried, but did not report
        # specific results for other than mentioning in the conclusion
        # that it helped.
        n_local_trials = 2 + int(np.log(n_clusters))

    # Pick first center randomly
    center_id = random.randint(n_samples)
    centers[0] = X[center_id]

    # Initialize list of closest distances and calculate current potential
    closest_dist_sq = euclidean_distances(
        centers[0], X, Y_norm_squared=x_squared_norms, squared=True)
    current_pot = closest_dist_sq.sum()

    # Pick the remaining n_clusters-1 points
    for c in range(1, n_clusters):
        # Choose center candidates by sampling with probability proportional
        # to the squared distance to the closest existing center
        rand_vals = random.random_sample(n_local_trials) * current_pot
        candidate_ids = np.searchsorted(closest_dist_sq.cumsum(), rand_vals)

        # Compute distances to center candidates
        distance_to_candidates = euclidean_distances(
            X[candidate_ids], X, Y_norm_squared=x_squared_norms, squared=True)

        # Decide which candidate is the best
        best_candidate = None
        best_pot = None
        best_dist_sq = None
        for trial in range(n_local_trials):
            # Compute potential when including center candidate
            new_dist_sq = np.minimum(closest_dist_sq,
                                     distance_to_candidates[trial])
            new_pot = new_dist_sq.sum()

            # Store result if it is the best local trial so far
            if (best_candidate is None) or (new_pot < best_pot):
                best_candidate = candidate_ids[trial]
                best_pot = new_pot
                best_dist_sq = new_dist_sq

        # Permanently add best center candidate found in local tries
        centers[c] = X[best_candidate]
        current_pot = best_pot
        closest_dist_sq = best_dist_sq

    return centers


def _tolerance(X, tol):
    """Return a tolerance which is independent of the dataset"""
    variances = np.var(X, axis=0)
    return np.mean(variances) * tol


class WeightedKMeans():
    """Weighted K-Means clustering

    Parameters
    ----------

    n_clusters : int, optional, default: 8
        The number of clusters to form as well as the number of
        centroids to generate.

    max_iter : int, default: 300
        Maximum number of iterations of the k-means algorithm for a
        single run.

    n_init : int, default: 10
        Number of time the k-means algorithm will be run with different
        centroid seeds. The final results will be the best output of
        n_init consecutive runs in terms of inertia.

    tol : float, default: 1e-4
        Relative tolerance with regards to inertia to declare convergence

    n_jobs : int, default: 1
        The number of jobs to use for the computation. This works by breaking
        down the pairwise matrix into n_jobs even slices and computing them in
        parallel.

        If -1 all CPUs are used. If 1 is given, no parallel computing code is
        used at all, which is useful for debugging. For n_jobs below -1,
        (n_cpus + 1 + n_jobs) are used. Thus for n_jobs = -2, all CPUs but one
        are used.


    Attributes
    ----------
    cluster_centers_ : array, [n_clusters, n_features]
        Coordinates of cluster centers

    labels_ :
        Labels of each point

    inertia_ : float
        Sum of distances of samples to their closest cluster center.

    Notes
    ------
    The k-means problem is solved using Lloyd's algorithm.

    The average complexity is given by O(k n T), were n is the number of
    samples and T is the number of iteration.

    The worst case complexity is given by O(n^(k+2/p)) with
    n = n_samples, p = n_features. (D. Arthur and S. Vassilvitskii,
    'How slow is the k-means method?' SoCG2006)

    In practice, the k-means algorithm is very fast (one of the fastest
    clustering algorithms available), but it falls in local minima. That's why
    it can be useful to restart it several times.

    """

    def __init__(self, n_clusters=8, n_init=10, max_iter=300, tol=1e-4):

        self.n_clusters = n_clusters
        self.max_iter = max_iter
        self.tol = tol
        self.n_init = n_init

    def _check_fitted(self):
        if not hasattr(self, "cluster_centers_"):
            raise AttributeError("Model has not been trained yet.")

    def fit(self, X, weights=None):
        """Compute k-means clustering.

        Parameters
        ----------
        X : array-like or sparse matrix, shape=(n_samples, n_features)
        """
        self.cluster_centers_, self.labels_, self.inertia_ = \
            weighted_kmeans(
                X, n_clusters=self.n_clusters, weights=weights,
                n_init=self.n_init, max_iter=self.max_iter, tol=self.tol)
        return self

    def fit_predict(self, X):
        """Compute cluster centers and predict cluster index for each sample.

        Convenience method; equivalent to calling fit(X) followed by
        predict(X).
        """
        return self.fit(X).labels_

    def fit_transform(self, X, y=None):
        """Compute clustering and transform X to cluster-distance space.

        Equivalent to fit(X).transform(X), but more efficiently implemented.
        """
        # Currently, this just skips a copy of the data if it is not in
        # np.array or CSR format already.
        # XXX This skips _check_test_data, which may change the dtype;
        # we should refactor the input validation.
        X = self._check_fit_data(X)
        return self.fit(X)._transform(X)

    def transform(self, X):
        """Transform X to a cluster-distance space.

        In the new space, each dimension is the distance to the cluster
        centers.  Note that even if X is sparse, the array returned by
        `transform` will typically be dense.

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape = [n_samples, n_features]
            New data to transform.

        Returns
        -------
        X_new : array, shape [n_samples, k]
            X transformed in the new space.
        """
        self._check_fitted()
        X = self._check_test_data(X)
        return self._transform(X)

    def _transform(self, X):
        """guts of transform method; no input validation"""
        return euclidean_distances(X, self.cluster_centers_)

    def predict(self, X):
        """Predict the closest cluster each sample in X belongs to.

        In the vector quantization literature, `cluster_centers_` is called
        the code book and each value returned by `predict` is the index of
        the closest code in the code book.

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape = [n_samples, n_features]
            New data to predict.

        Returns
        -------
        labels : array, shape [n_samples,]
            Index of the cluster each sample belongs to.
        """
        self._check_fitted()
        x_squared_norms = np.einsum('ij,ij->i', X, X)
        return _labels_inertia(X, x_squared_norms, self.cluster_centers_)[0]

    def score(self, X):
        """Opposite of the value of X on the K-means objective.

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape = [n_samples, n_features]
            New data.

        Returns
        -------
        score : float
            Opposite of the value of X on the K-means objective.
        """
        self._check_fitted()
        x_squared_norms = np.einsum('ij,ij->i', X, X)
        return -_labels_inertia(X, x_squared_norms, self.cluster_centers_)[1]



__author__ = 'jiohyoo'

import numpy as np
from numpy.linalg import inv
from numpy.random import multivariate_normal
from sklearn.mixture import GMM
from sklearn.mixture.gmm import _log_multivariate_normal_density_full
from sparse_gmm import SparseGMM


def predict_missing_values(model, data):
    if not isinstance(model, GMM):
        raise ValueError('The given model is not an instance of GMM or SparseGMM')

    predicted_data = data.copy()

    if len(predicted_data.shape) == 1:
        predicted_data = predicted_data[:, np.newaxis]

    n_frames, n_vars = predicted_data.shape
    for idx_frame in range(n_frames):
        row = predicted_data[idx_frame:idx_frame+1]
        idx_missing = np.isnan(row[0])
        c, mu, sigma_none = get_conditional_dist(model, row, skip_covar=True)
        predicted_data[idx_frame][idx_missing] = mu
        if model.nonparanormal:
            predicted_data[idx_frame] = \
                model.npns_[c].inverse_transform(predicted_data[idx_frame:idx_frame+1])

    return predicted_data

def sample_missing_values(model, data, n_samples=10):
    if not isinstance(model, GMM):
        raise ValueError('The given model is not an instance of GMM or SparseGMM')

    if len(data.shape) == 1:
        data = data[:, np.newaxis]
    n_frames, n_vars = data.shape
    predicted_data = np.empty((n_frames * n_samples, n_vars))

    for idx_frame in range(n_frames):
        row = data[idx_frame:idx_frame+1]
        c, mu, sigma = get_conditional_dist(model, row, skip_covar=False)

        idx_missing = np.isnan(row[0])

        for i in range(n_samples):
            if len(mu.shape) == 0:
                mu = mu.reshape((1,))
            sampled_vals = multivariate_normal(mu, sigma)
            predicted_data[idx_frame * n_samples + i] = row
            predicted_data[idx_frame * n_samples + i][idx_missing] = sampled_vals
            if model.nonparanormal:
                predicted_data[idx_frame * n_samples + i] = \
                    model.npns_[c].inverse_transform(
                        predicted_data[idx_frame * n_samples + i : idx_frame * n_samples + i + 1])

    return predicted_data


def get_conditional_dist(model, row, skip_covar=True):
    if not isinstance(model, GMM):
        raise ValueError('The given model is not an instance of GMM or SparseGMM')

    row = np.asarray(row, dtype='float')
    if len(row.shape) == 1:
        row = row[np.newaxis, :]

    n_vars = row.shape[1]

    idx_given = []
    idx_missing = []
    for i in range(n_vars):
        if np.isnan(row[0, i]):
            idx_missing.append(i)
        else:
            idx_given.append(i)

    if len(idx_given) == 0:
        raise ValueError('Row must have at least one given value!')
    if len(idx_missing) == 0:
        raise ValueError('The given row has all values filled already!')

    # prepare parts of the params
    mu_g = model.means_[:, idx_given]
    mu_m = model.means_[:, idx_missing]

    sigma_gg = model.covars_[np.ix_(range(model.n_components), idx_given, idx_given)]
    sigma_mg = model.covars_[np.ix_(range(model.n_components), idx_missing, idx_given)]
    sigma_mm = model.covars_[np.ix_(range(model.n_components), idx_missing, idx_missing)]

    # marginal probability for each component
    if isinstance(model, SparseGMM) and model.nonparanormal:
        marginal_probs = np.empty(model.n_components)
        for c in range(model.n_components):
            row_t = model.npns_[c].transform(row)
            vals_given = row_t[0, np.ix_(idx_given)]
            marginal_probs[c] = np.exp(_log_multivariate_normal_density_full(vals_given, mu_g[c:c+1], sigma_gg[c:c+1]))
    else:
        vals_given = row[0, np.ix_(idx_given)]
        marginal_probs = np.exp(_log_multivariate_normal_density_full(vals_given, mu_g, sigma_gg))

    # weighted marginal probability for each component
    probs = marginal_probs * model.weights_

    # find the most probable component
    k = np.argmax(probs)

    # conditional distribution
    mu_c = mu_m[k] + sigma_mg[k].dot(inv(sigma_gg[k])).dot(vals_given[0] - mu_g[k])[0]
    if skip_covar:
        return k, mu_c, None
    else:
        sigma_c = sigma_mm[k] + sigma_mg[k].dot(inv(sigma_gg[k])).dot(sigma_mg[k].T)
        return k, mu_c, sigma_c




import numpy as np
from scipy.stats import norm

class NPNTransformer():
    """
    Transformation function for multi-variate non-paranormal data

    self.vals[j][idx] = j-th dimension, val of the sample at index idx when sorted
    self.cnts[j][idx] = j-th dimension, cdf at index idx
    self.mu[j] = sample mean of j-th dimension data
    self.sigma[j] = sample std of j-th dimension data

    self.delta = if not specified, follow Liu (2009) The Nonparanormal paper
    """
    def __init__(self):
        self.is_data_set = False
        self.n_dims = 0
        self.n_samples = 0
        self.cdf = None
        self.vals = None
        self.delta = 0
        self.mu = None
        self.sigma = None
        self.ratio = None

    def set_data(self, data, delta=None, ratio=1.0):
        if self.is_data_set is True:
            raise RuntimeError('Data is already set!')

        if len(data.shape) == 1:
            data = data.reshape((data.size, 1))

        if ratio <= 0 or ratio > 1.0:
            raise ValueError('The sample ratio for NPNTransformer %f must be beween 0 and 1!' % ratio)
        self.ratio = ratio

        n_samples = int(round(data.shape[0] * ratio))
        if n_samples != data.shape[0]:
            data = data[np.random.permutation(data.shape[0])[:n_samples]]
        else:
            data = data.copy()

        self.n_samples, self.n_dims = data.shape

        if delta is None:
            self.delta = 1.0 / (4 * (self.n_samples ** 0.25) * np.sqrt(np.pi * np.log(self.n_samples)))
        else:
            self.delta = delta

        self.cdf = [None] * self.n_dims
        self.vals = [None] * self.n_dims
        for dim in range(self.n_dims):
            sorted_col = np.sort(data[:, dim])
            n_unique = np.unique(data[:, dim]).size
            self.cdf[dim] = np.zeros(n_unique, dtype=float)
            self.vals[dim] = np.zeros(n_unique, dtype=float)

            prev_val = None
            prev_idx = -1
            for i in range(sorted_col.size):
                if sorted_col[i] == prev_val:
                    self.cdf[dim][prev_idx] += 1
                else:
                    self.cdf[dim][prev_idx + 1] = self.cdf[dim][prev_idx] + 1
                    self.vals[dim][prev_idx + 1] = sorted_col[i]
                    prev_val = sorted_col[i]
                    prev_idx += 1
            self.cdf[dim] /= self.n_samples

        self.mu = data.mean(axis=0)
        self.sigma = data.std(axis=0)
        self.is_data_set = True

    def transform(self, X):
        if self.is_data_set is False:
            raise RuntimeError('Data is not set yet!')

        if len(X.shape) == 1:
            Xt = np.zeros((X.size, 1), dtype=float)
            X = X.reshape((X.size, 1))
        else:
            Xt = np.zeros(X.shape, dtype=float)

        n_samples, n_dims = Xt.shape
        if n_dims != self.n_dims:
            raise ValueError('dimension of the given X %d is not equal to the transformation\'s dimension %d' %
                             (n_dims, self.n_dims))

        # get empirical distribution
        for sample_idx in range(n_samples):
            for dim in range(n_dims):
                v = X[sample_idx, dim]
                if np.isnan(v):
                    Xt[sample_idx, dim] = np.nan
                else:
                    if v < self.vals[dim][0]:
                        Xt[sample_idx, dim] = 0.
                    elif v >= self.vals[dim][-1]:
                        Xt[sample_idx, dim] = 1.
                    else:
                        max = self.vals[dim].size - 1
                        min = 0
                        mid = -1
                        while min <= max:
                            mid = (max + min) / 2
                            if self.vals[dim][mid] > v:
                                max = mid - 1
                            elif self.vals[dim][mid] < v:
                                min = mid + 1
                            else:
                                break
                        if self.vals[dim][mid] > v:
                            Xt[sample_idx, dim] = self.cdf[dim][mid - 1]
                        else:
                            Xt[sample_idx, dim] = self.cdf[dim][mid]

        # winsorization
        Xt[np.logical_and(np.isnan(Xt) == False, Xt < self.delta)] = self.delta
        Xt[np.logical_and(np.isnan(Xt) == False, Xt > 1 - self.delta)] = 1 - self.delta

        # apply inverse CDF
        Xt = norm.ppf(Xt)

        # apply mean and std
        Xt = Xt * self.sigma + self.mu
        return Xt

    def inverse_transform(self, Xt):

        if self.is_data_set is False:
            raise RuntimeError('Data is not set yet!')

        X = (Xt - self.mu) / self.sigma
        X = norm.cdf(X)

        n_samples, n_dims = Xt.shape
        if n_dims != self.n_dims:
            raise ValueError('dimension of the given X %d is not equal to the transformation\'s dimension %d' %
                             (n_dims, self.n_dims))

        # get the reverse of the empirical distribution
        for sample_idx in range(n_samples):
            for dim in range(n_dims):
                v = X[sample_idx, dim]
                if not np.isnan(v):
                    # if v < self.vals[dim][0]:
                    #     Xt[sample_idx, dim] = 0.
                    # elif v >= self.vals[dim][-1]:
                    #     Xt[sample_idx, dim] = 1.
                    # else:
                    max = self.cdf[dim].size - 1
                    min = 0
                    mid = -1
                    while min <= max:
                        mid = (max + min) / 2
                        if self.cdf[dim][mid] > v:
                            max = mid - 1
                        elif self.cdf[dim][mid] < v:
                            min = mid + 1
                        else:
                            break
                    if self.cdf[dim][mid] > v:
                        X[sample_idx, dim] = self.vals[dim][mid - 1]
                    else:
                        X[sample_idx, dim] = self.vals[dim][mid]

        return X




import base64
import json
import numpy as np
from sparse_gmm import SparseGMM
from nonparanormal import NPNTransformer

class GamelanPyEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, SparseGMM):
            dct = obj.__dict__
            dct['__sparse_gmm__'] = True
            return dct
        if isinstance(obj, NPNTransformer):
            print 'hello'
            dct = obj.__dict__
            dct['__npn_transformer__'] = True
            return dct
        if isinstance(obj, np.ndarray):
            data_base64 = base64.b64encode(obj.data)
            return dict(__ndarray__=data_base64, dtype=str(obj.dtype), shape=obj.shape)

        return json.JSONEncoder(self, obj)

def gamelan_json_obj_hook(dct):
    if isinstance(dct, dict) and '__ndarray__' in dct:
        data = base64.b64decode(dct['__ndarray__'])
        return np.frombuffer(data, dct['dtype']).reshape(dct['shape'])
    if isinstance(dct, dict) and '__sparse_gmm__' in dct:
        sgmm = SparseGMM()
        sgmm.__dict__ = dct
        return sgmm
    if isinstance(dct, dict) and '__npn_transformer__' in dct:
        npn = NPNTransformer()
        npn.__dict__ = dct
        return npn
    return dct



#!/usr/bin/python
__author__ = 'jiohyoo'
import sys
import gamelanpy
import getopt
import numpy as np
import json

def getopts(argv):
    try:
        opts, args = getopt.getopt(argv, "h", [
            "help", "subsample-method=", "subsample-size=",
            "save-model=", "save-weights=", "save-means=", "save-covars=", "save-precs=",
            "l1-depth=", "l1-min=", "l1-max=", "l1-search-repeat=",
            "npn-sample-ratio="])
    except getopt.GetoptError:
        usage()
        raise
        exit(2)

    fit_options = {}
    save_options = {}
    npn = False
    npn_ratio = 0.0

    fit_options['subsample_size'] = None
    for opt, arg in opts:
        print opt, arg
        if opt in ['-h', '--help']:
            usage()
            exit()
        elif opt == '--save-model':
            save_options['model'] = arg
        elif opt == '--save-covars':
            save_options['covars'] = arg
        elif opt == '--save-means':
            save_options['means'] = arg
        elif opt == '--save-precs':
            save_options['precs'] = arg
        elif opt == '--save-weights':
            save_options['weights'] = arg
        elif opt == '--l1-depth':
            fit_options['l1_search_depth'] = int(arg)
        elif opt == '--l1-min':
            fit_options['l1_penalty_range_min'] = float(arg)
        elif opt == '--l1-max':
            fit_options['l1_penalty_range_max'] = float(arg)
        elif opt == '--subsample-method':
            fit_options['subsample_method'] = arg
        elif opt == '--subsample-size':
            fit_options['subsample_size'] = int(arg)
        elif opt == '--l1-search-repeat':
            fit_options['l1_search_repeat'] = int(arg)
        elif opt == '--npn-sample-ratio':
            npn = True
            fit_options['npn_sample_ratio'] = float(arg)

    if 'subsample_method' not in fit_options:
        fit_options['subsample_method'] = 'None'
        print 'Warning: Subsampling is not used for learning process.'

    if 'subsample_method' in fit_options \
            and fit_options['subsample_method'] in ['coreset', 'coreset2', 'uniform'] \
            and 'subsample_size' not in fit_options:
        print 'Subsampling method %s is used, but subsample_size is not specified!' % fit_options['subsample_method']
        print 'Please specify the subsample size.'
        exit(2)

    if ('l1_penalty_range_min' in fit_options) is not ('l1_penalty_range_max' in fit_options):
        print 'Please specify both of l1_penalty min and max values.'
        print 'To use default values, do not specify both min and max values.'
        exit(2)

    if npn:
        if fit_options['npn_sample_ratio'] <= 0 or fit_options['npn_sample_ratio'] > 1:
            print 'Nonparanormal sample ratio: %f must be between 0 and 1' % fit_options['npn_sample_ratio']
            exit(2)

    if 'model' not in save_options:
        print 'WARNING: the whole representation of the learned model will not be saved.'

    return fit_options, save_options, npn


def main(data_path, n_components, fit_options, save_options, npn):
    sgmm = gamelanpy.SparseGMM(n_components=n_components, nonparanormal=npn)
    data = np.genfromtxt(data_path, delimiter=',')
    sgmm.fit(data, **fit_options)

    n_vars = sgmm.means_.shape[1]
    if 'model' in save_options:
        json.dump(sgmm, open(save_options['model'], 'w'), cls=gamelanpy.GamelanPyEncoder)
    if 'covars' in save_options:
        np.savetxt(save_options['covars'], sgmm.covars_.reshape((n_components * n_vars, n_vars)), delimiter=',')
    if 'means' in save_options:
        np.savetxt(save_options['means'], sgmm.means_, delimiter=',')
    if 'weights' in save_options:
        np.savetxt(save_options['weights'], sgmm.weights_, delimiter=',')
    if 'precs' in save_options:
        np.savetxt(save_options['precs'], sgmm.precs_.reshape((n_components * n_vars, n_vars)), delimiter=',')


def usage():
    print 'print usage'

if __name__ == '__main__':
    n_components = int(sys.argv[1])
    data_path = sys.argv[2]
    fit_options, save_options, npn = getopts(sys.argv[3:])
    main(data_path, n_components, fit_options, save_options, npn)






import logging
import numpy as np
import scipy.linalg
import scipy.cluster
import scipy.spatial
import time


def get_coreset(data, num_clusters, coreset_size, delta=0.1):
    '''
    Parameters
    ----------
    data: array-like, (num_frames, num_vars)
    num_clusters: int, number of clusters
    coreset_size: int, number of coreset samples
    delta: float, default=0.1

    Return
    ------
    samples: coreset samples
    weights: coreset weights
    '''

    logger = logging.getLogger()

    if len(data.shape) == 1:
        logger.debug('Input data is 1-D, converting it to 2-D')
        data = data[:, np.newaxis]

    num_frames, num_vars = data.shape

    if coreset_size < num_clusters:
        raise ValueError("coreset_size %d is less than num_mixtures %d" % (coreset_size, num_clusters))

    if num_frames < coreset_size:
        raise ValueError("num_frames %d is less than coreset_size %d" % (num_frames, num_clusters))

    data_remain = data.copy()

    samples = np.zeros((0, num_vars))
    # first, do the subsampling : pick core samples, and remove closest point to
    # it

    logger.debug('Before Coreset random sampling')

    num_iters = 0
    num_single_samples = int(1.0 * num_vars * num_clusters * np.log(1.0 / delta))
    logger.debug('num_single_samples: %d', num_single_samples)

    while data_remain.shape[0] > num_single_samples:
        cur_time = time.time()
        logger.debug('Starting iteration %d', num_iters)

        num_frames_remain = data_remain.shape[0]
        idx = np.random.permutation(num_frames_remain)[:num_single_samples]
        single_samples = data_remain[idx, :]

        prev_time = cur_time
        cur_time = time.time()
        logger.debug('After random sampling (took %.3f sec)', cur_time - prev_time)

        # Here we define similarity matrix, based on some measure of
        # similarity or kernel. Feel free to change

        dists = scipy.spatial.distance.cdist(data_remain, single_samples)

        prev_time = cur_time
        cur_time = time.time()
        logger.debug('After evaluating cdist (took %.3f sec)', cur_time - prev_time)

        # minimum distance from random samples
        min_dists = np.min(dists, axis=1)
        # median distance
        v = np.median(min_dists)

        # remove rows with distance <= median distance
        remove_idx = np.where(min_dists <= v)[0]

        # remove rows of remove_idx
        data_remain = np.delete(data_remain, remove_idx, 0)
        samples = np.vstack((samples, single_samples))
        logger.debug('Shape of the coreset samples so far (%d, %d)', samples.shape)
        logger.debug('Shape of the remaining samples (%d, %d)', data_remain.shape)

        prev_time = cur_time
        cur_time = time.time()
        logger.debug('End of iteration %d (took %.3f sec)', (num_iters, cur_time - prev_time))

        num_iters += 1
    # end of while loop

    logger.debug('Shape of the final remaining samples (%d, %d)', data_remain.shape)

    samples = np.vstack((samples, data_remain))

    logger.debug('Shape of the final coreset samples (%d, %d)', samples.shape)

    # now compute the weights of all the points, according to how close they
    # are to the closest core-sample.
    db_size = np.zeros(samples.shape[0])
    min_dists = np.zeros(num_frames)
    closest_sample_idx = np.zeros(num_frames)
    for i in xrange(num_frames):
        dists = scipy.spatial.distance.cdist(data[i:i+1, :], samples)
        min_dist = np.min(dists)
        min_idx = np.argmin(dists)
        min_dists[i] = min_dist
        closest_sample_idx[i] = min_idx

    for i in xrange(num_frames):
        # for each datapoint, Ix[i] is the index of the coreset point
        # it is assigned to.
        db_size[closest_sample_idx[i]] += 1

    sq_sum_min_dists = (min_dists ** 2).sum()
    m = np.zeros(num_frames)
    for i in xrange(num_frames):
        m[i] = np.ceil(5.0 / db_size[closest_sample_idx[i]] + (min_dists[i] ** 2) / sq_sum_min_dists)

    m_sum = m.sum()
    cdf = (1.0 * m / m_sum).cumsum()
    samples = np.zeros((coreset_size, num_vars))
    weights = np.zeros(coreset_size)

    # Now, sample from the weighted points, to generate final corset
    # and the corresponding weights
    for i in xrange(coreset_size):
        r = np.random.rand()
        idx = (cdf <= r).sum()
        samples[i, :] = data[idx, :]
        weights[i] = m_sum / (coreset_size * m[idx])

    return samples, weights






from distutils.core import setup, Extension
import numpy as np

ext_weighted_kmeans = Extension('_weighted_kmeans',
                                sources=['gamelanpy/_weighted_kmeans.c'],
                                include_dirs=[np.get_include()])

setup(
    name='gamelanpy',
    version='0.1.2',
    author='Ji Oh Yoo',
    author_email='jioh.yoo@gmail.com',
    packages=['gamelanpy'],
    scripts=['scripts/gamelan_learn_model.py', 'scripts/gamelan_imputation.py'],
    license='LICENSE.txt',
    description='Python Implementation of GAMELAN core',
    install_requires=[
        "numpy >= 1.6.0",
        "scipy >= 0.13.0",
        "scikit-learn >= 0.16.0"
    ],
    ext_modules=[ext_weighted_kmeans]
)